# -*- coding: utf-8 -*-
from openkongqi.source import get_source


def source_router(name, args, kwargs, options, task=None):
    """
    A dynamic router used to set a specific queue to scrape a source if the
    source definition has a key named 'queue'. If no queue is specified, return
    `None` which will set the queue to `task_default_queue`.

    .. code:: json

        {
          "taiwan": {
            "target": "https://taqm.epa.gov.tw/pm25/en/PM25A.aspx?area=10",
            "uuid": "tw",
            "modname": "okq_gams.source.taqm",
            "tz": "Asia/Taipei",
            "queue": "taiwan"
          }
        }
        
    """
    if name == 'openkongqi.tasks.scrape':
        info = get_source(args[0])
        if 'queue' in info:
            return {
                'queue': info['queue'],
                'routing_key': info['queue'],
            }
        else:
            return None



# -*- coding: utf-8 -*-

import redis

from .base import BaseCacheWrapper

# default redis settings
_HOST = 'localhost'
_PORT = '6379'
_DB_ID = 0


class CacheWrapper(BaseCacheWrapper):

    def create_cnx(self, db_settings):
        return redis.StrictRedis(
            host=db_settings.get('HOST', _HOST),
            port=db_settings.get('PORT', _PORT),
            db=db_settings.get('DB_ID', _DB_ID),
            charset="utf-8",
            decode_responses=True
        )

    def set(self, key, value):
        return self._cnx.set(key, value)

    def get(self, key):
        return self._cnx.get(key)



# -*- coding: utf-8 -*-

from .utils import get_source

from celery import Celery
from celery.utils.log import get_task_logger


app = Celery('okq')
app.config_from_object('celeryconfig')
logger = get_task_logger(__name__)


@app.task
def scrape(name):
    src = get_source(name)
    src.scrape()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from datetime import datetime
from inspect import getmodulename
import logging
import os
import re
import socket
import ssl
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, Request

from ..conf import settings, logger, statusdb, recsdb, file_cache
from ..exceptions import SourceError
from ..stations import get_station_map
from ..apikeys import get_api_key
from ..utils import get_rnd_item, get_uuid

import pytz


HTTP_TIMEOUT = 10

logger = logging.getLogger("openkongqi.source.{}"
                           .format(getmodulename(__file__)))


class BaseSource(object):
    """Base source class to scrape online resources. This class is to be used
    as parent of any scraping class.

    Basic mechanism from fetching online resource to hosing extracted data in
    the database, :meth:`openkongqi.source.BaseSource.scrape` is the main entry
    point of this class.
    """

    key_context = None
    _now = None

    def __init__(self, name):
        """
        :param name: fetch name
        :type name: str
        """
        if name not in settings['SOURCES']:
            raise SourceError('Unknown source ({})'.format(name))
        self.name = name
        self.target = settings['SOURCES'][name]['target']
        self._station_map = get_station_map(
            settings['SOURCES'][name]['uuid'])
        self._api_key = get_api_key(name.split(":")[0])
        self._tz = pytz.timezone(settings['SOURCES'][name]['tz'])
        self._status = statusdb
        self._cache = file_cache
        self._records = recsdb

    def scrape(self):
        """Main entry point for :class:`openkongqi.source.BaseSource` instances.

        Orchestrates the resource scraping, following actions are performed:

        * :meth:`openkongqi.source.BaseSource.fetch`: fetch online resource
        * :meth:`openkongqi.source.BaseSource.save_status`: save fetching
          status
        * :meth:`openkongqi.source.BaseSource.cache`: save the resource in the
          cache
        * :meth:`openkongqi.source.BaseSource.extract`: extract the data
          from the resource
        * :meth:`openkongqi.source.BaseSource.save_data`: save extracted data
        """
        self._now = datetime.now(pytz.utc)
        src_content = self.fetch()
        self.save_status()
        if src_content is not None:
            content = self.cache(src_content)
            data = self.extract(content)
            self.save_data(data)

    def fetch(self):
        """Fetch the resource

        .. warning:: This method has to be overwritten

        :returns: content - a file-like object
        """
        raise NotImplementedError

    def post_fetch(self, resource):
        """Treat the resource after fetching raw content

        .. warning:: This method has to be overwritten

        :param resource: raw content / data (could be http, csv, xml, etc...)
        :returns: content - a file-like object
        """
        raise NotImplementedError

    def get_station_uuid(self, name):
        """Return the uuid of a station

        .. note:: This is not performing any recursive search in the map
        """
        return get_uuid(settings['SOURCES'][self.name]['uuid'],
                        self._station_map[name]['uuid'])

    def get_status_data(self):
        """Get the fetch status

        :returns: data - a dict with data to serialize in the status entry
        """
        return dict()

    def save_status(self):
        """Save status data to keep track of fetching history.

        Uses :meth:`openkongqi.source.BaseSource.get_status_data` to get the
        status data to save.
        """
        data = self.get_status_data()
        if data is None:
            data = {'ts': self._now.strftime('%Y%m%d%H%M%S')}
        else:
            data.setdefault('ts', self._now.strftime('%Y%m%d%H%M%S'))
        self._status.set_status(self.name, data)

    def cache(self, content):
        """Cache fetching content

        :param content: content to cache
        :type content: file-like object
        """
        self._cache.set(self.name, content, self._now)
        fp = self._cache.get_fp(self.name, self._now)
        # display how much is cached into server
        logger.info("Cached {} kilobytes to server."
                    .format(os.path.getsize(fp)))
        return open(fp, 'rb')

    def pythonify(self, text, is_num=False):
        if text is None:
            return None

        if self.null_re is not None:
            if self.null_re.match(text) is None:
                if is_num:  # remove everything but nums and dots
                    return float(re.sub(r'[^\d.]+', '', text))
                else:
                    return text
            else:
                return None
        else:
            if is_num:  # remove everything but nums and dots
                try:
                    return float(re.sub(r'[^\d.]+', '', text))
                except ValueError:
                    pass
            else:
                return text

    def extract(self, content):
        """Extract data from the content
        """
        raise NotImplementedError

    def save_data(self, data, ignore_check_latest=False):
        self._records.write_records(data,
                                    ignore_check_latest=ignore_check_latest,
                                    context=self.key_context)

    def get_latest(self, uuid):
        return self._records.get_latest(uuid, context=self.key_context)

    def get_records(self, uuid, start, end, filters=None):
        return self._records.get_records(uuid=uuid,
                                         start=start,
                                         end=end,
                                         context=self.key_context)


class HTTPSource(BaseSource):

    def __init__(self, name):
        super(HTTPSource, self).__init__(name)

    def fetch(self):

        _common_error_header = "Data fetch failure"

        try:
            req = self.get_req(self.target)
            resp = urlopen(req, timeout=HTTP_TIMEOUT)
        except HTTPError as e:
            self._info = None
            self._statuscode = e.code
            logger.error("{}: {} {} ({})"
                         .format(_common_error_header,
                                 e.code, e.msg, e.url))
            return None
        except URLError as e:
            if isinstance(e.reason, socket.timeout):
                self._info = None
                self._statuscode = 418
                logger.error("{}: HTTP Timeout".format(_common_error_header))
                return None
            else:
                self._info = None
                self._statuscode = None
                logger.error("{}: {}"
                             .format(_common_error_header, e.reason.strerror))
                return None
        except socket.timeout:
            self._info = None
            self._statuscode = 418
            logger.error("{}: HTTP Timeout".format(_common_error_header))
            return None
        except ssl.CertificateError as e:
            logger.error("{}: {}"
                         .format(_common_error_header, e.message))
            return None

        self._info = resp.info()
        self._statuscode = resp.getcode()
        logger.debug("Fetch status: HTTP {}".format(self._statuscode))

        if self._statuscode == 204:  # 204: No Content
            logger.warning("No message body included")
            return None

        try:
            # don't use `resp.headers.get('content-length`)
            # because if it doesn't exist in headers it will return None
            # which makes it a false negative
            content_length = resp.headers['content-length']
        except KeyError:
            pass
        else:
            if (self._statuscode == 200 and content_length == '0'):
                logger.warning("Fetched content is empty; skipping cache ...")
                return None

        return self.post_fetch(resp)

    def post_fetch(self, resource):
        return resource

    def get_req(self, target):
        """Return an :class:`urllib2.Request` instance
        """
        req = Request(target)
        req.add_header('User-Agent', get_rnd_item(settings['UA_FILE']))
        return req

    def get_status_data(self):
        data = {
            'code': self._statuscode
        }
        if self._info is not None and 'Last-Modified' in self._info:
            data['last-modified'] = self._info['Last-Modified']
        return data



# -*- coding: utf-8 -*-
from datetime import datetime
import json
import pytz

from ..utils import load_backend

_CACHE_KEY = 'okq:{moduuid}:{uuid}:latest'


class BaseRecordsWrapper(object):
    """Base wrapper class to get database records. This class is to be used
    as parent of any database records wrapper class.

    The children wrapper class will have to define custom methods,
    for creating the database connection.
    """

    ts_fmt = "%Y-%m-%dT%H:%M:%SZ"

    def __init__(self, settings, cache, *args, **kwargs):
        self._cnx = self.create_cnx(settings)
        self._cache = cache
        # NOTE: can't apply key context here
        # because it hasn't been set when this is initialized
        self._cache_key = settings.get('CACHE_KEY', _CACHE_KEY)

    def create_cnx(self, settings):
        """Create a connection to the database

        .. warning:: This method has be overwritten
        """
        raise NotImplementedError

    def db_init(self):
        """Initialize database.

        .. warning:: This method has be overwritten
        """
        raise NotImplementedError

    def is_duplicate(self, record):
        """Check for duplicated records.

        Check if the timestamp, uuid, key already exists in the db.
        Note that these are all the primary keys.

        .. warning:: This method has be overwritten
        """
        raise NotImplementedError

    def write_records(self, records, ignore_check_latest=False, context=None):
        """Save the records

        See data extraction format on what to expect as input.

        .. warning:: This method has to be overwritten
        """
        raise NotImplementedError

    def get_records(self, start, end, filters=None, context=None):
        """Returns a list of Records.

        .. warning:: This method has to be overwritten

        :param start: the start date (lower boundary)
        :param end: the end date (upper boundary)
        :param filters: list of columns to select
        :type filters: list of str
        """
        raise NotImplementedError

    def _get_cache_key(self, uuid, context=None):
        ctx_fmt = {u'uuid': uuid}
        if context is not None:
            ctx_fmt.update(context)
        return self._cache_key.format(**ctx_fmt)

    def set_latest(self, uuid, record, context=None):
        """Set record as the latest entry in cache database.

        :param uuid: unique id
        :type uuid: str
        """
        latest_record = {
            'ts': self._ts_to_string(record['ts']),
            'fields': record['fields'],
        }
        latest_json = json.dumps(latest_record)
        key = self._get_cache_key(uuid=uuid, context=context)
        self._cache.set(key, latest_json)

    def get_latest(self, uuid, context=None):
        """Get latest record entry from cache database.

        :param uuid: unique id
        :type uuid: str
        """
        latest = self._cache.get(
            self._get_cache_key(uuid=uuid, context=context)
        )
        if latest is None:
            return None
        record = json.loads(latest)
        return {
            'ts': self._string_to_ts(record['ts']),
            'fields': record['fields'],
        }

    def _ts_to_string(self, ts):
        """Convert a datetime.datetime object to a string.

        This is done for compatibility in serialization.

        Format:    %Y-%m-%dT%H:%M:%SZ
        Example: 2016-07-13T10:09:56Z

        :param ts: timestamp
        :type ts: datetime.datetime
        """
        # remove the microseconds
        ts = ts.replace(microsecond=0)
        # convert to UTC or keep naive
        if ts.tzinfo is not None:
            ts = ts.astimezone(pytz.utc)
        return datetime.strftime(ts, self.ts_fmt)

    def _string_to_ts(self, ts_string):
        """Convert a timestamp string to datetime.datetime object.

        The input format has to be the same
        from the output of ``_ts_to_string``.

        :param ts_string: timestamp string
        :type ts_string: str
        """
        return datetime \
            .strptime(ts_string, self.ts_fmt) \
            .replace(tzinfo=pytz.utc)


def create_recsdb(settings, cache):
    mod = load_backend(settings['ENGINE'])
    return mod.RecordsWrapper(settings, cache)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals
from copy import deepcopy

from .conf import settings
from .utils import get_uuid, load_tree, passthrough_loader, SEP, WILDCARD


_STATIONS_MAP = load_tree(settings['STATIONS_MAP_DIR'],
                          data_loader=passthrough_loader)


def get_station_map(uuid=None):
    """Get the entire station map given a UUID, or get the concatenated station
    map from multiple cities/regions given a UIID with a wildcard "*"

    Usage::

        >>> from openkongqi.stations import get_station_map
        >>> get_station_map('th') # get the station map of Thailand
        >>> get_station_map('th:*') # get the station map of Thailand
        >>> get_station_map('cn:guangdong') # get the station map of Guangdong province
        >>> get_station_map('cn:*') # get the concatenated station map of all provinces in China

    :param uuid: a UUID key
    :type uuid: str
    """
    if uuid.endswith(SEP + WILDCARD) is True:
        id_map = {}
        uuid_without_wildcard = uuid[:-2]
        for k in _STATIONS_MAP:
            if k.startswith(uuid_without_wildcard):
                if uuid_without_wildcard == k:
                    id_map.update(_STATIONS_MAP.get(k, {}))
                else:
                    prefix = k.replace(uuid_without_wildcard + SEP, "", 1)
                    new_map = _inject_uuid_prefix(prefix, _STATIONS_MAP[k])
                    id_map.update(new_map)

    else:
        id_map = _STATIONS_MAP.get(uuid, {})
    return id_map


def get_all_uuids():
    """Return a sorted list of complete station UUIDs."""
    res = []
    for country, station_map in _STATIONS_MAP.items():
        res.extend([get_uuid(country, s) for s in station_map.values()])
    return sorted(res)


def _inject_uuid_prefix(prefix, station_map):
    map_copy = deepcopy(station_map)
    for station in map_copy:
        map_copy[station]["uuid"] = get_uuid(prefix, map_copy[station]["uuid"])
    return map_copy



# -*- coding: utf-8 -*-
from codecs import open
from os import path
from setuptools import setup, find_packages

import openkongqi as okq


# local path
here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst')) as fd:
    long_description = fd.read()

requirements = [
    "beautifulsoup4==4.9.3",
    "celery>=5.0.5",
    "hiredis==1.1.0",
    "html5lib==1.1",
    "pytz>=2020.5",
    "redis==3.5.3",
    "six>=1.13.0",
    "sqlalchemy>=1.3.23",
]

setup(
    name=okq.__name__,
    version=okq.__version__,
    author=okq.__author__,
    author_email=okq.__contact__,
    license="Apache License 2.0",
    packages=find_packages(exclude=['docs', 'test*']),
    url="https://github.com/gams/openkongqi",
    description="Outdoor air quality data",
    long_description=long_description,
    install_requires=requirements,
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'Framework :: Sphinx',
        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: Science/Research',
        'Natural Language :: Chinese (Simplified)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Text Processing :: Markup :: XML',
    ],
    keywords="air quality",
    package_data={
        'openkongqi': [
            'data/user_agent_strings.json',
            'data/sources/pm25.in.json',
            'openkongqi/data/stations/cn/shanghai.json'
        ],
    },
    entry_points={
        'console_scripts': [
            "okq-server=openkongqi.bin:okq_server",
            "okq-init=openkongqi.bin:okq_init",
            "okq-source-test=utils.source_test:main",
        ]
    },
    include_package_data=True,
)



# -*- coding: utf-8 -*-

from ..utils import load_backend


class BaseCacheWrapper(object):
    """A naive cache implementation with no data invalidation"""

    def __init__(self, db_settings, *args, **kwargs):
        self._cnx = self.create_cnx(db_settings)

    def create_cnx(self, db_settings):
        """Create a connection to the database.

        .. warning:: This method has to be overwritten
        """
        raise NotImplementedError

    def set(self, key, value):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError


def create_cachedb(settings):
    mod = load_backend(settings['ENGINE'])
    return mod.CacheWrapper(settings)
